
open Misc
open Config
open Format
open Typedtree
open Compenv

(* Option *)

let with_labels = ref false

open Lambda

let print_code x =
  if !with_labels
  then GccPrint.print_label_arg_program x
  else GccPrint.print_arg_program x

let dump_gcc outputprefix gcc =
  let objfile = outputprefix ^ ".gcc" in
  let oc = open_out_bin objfile in
  print_code oc gcc;
  close_out oc

let ppf = Format.err_formatter
let arg_list = Arg.align [
    "-interp", Arg.Clear GccComp.machine,
    "generate code for the interpret";
    "-dlambda", Arg.Set GccComp.print_lambda,
    "print intermediate lambda";
    "-label", Arg.Set with_labels,
    "generate code with labels";
    "-comment", Arg.Set GccPrint.print_comment,
    "generate code with comment";
    "-tail", Arg.Set GccGccgen.compile_tailcalls,
    "optimize tailcalls";
  ]

let files_to_concat = ref []

let arg_usage = Printf.sprintf "%s FILE.ml : compile to .gcc" Sys.argv.(0)

let arg_anon s =
  try
    if Filename.check_suffix s ".ml" then
      files_to_concat := s :: !files_to_concat
    else begin
      Printf.eprintf "Error: don't know what to do with %S\n%!" s;
      prerr_endline "Usage: ./ocamlgcc [options] file1.ml [file2.ml...]";
      exit 2
    end
  with x ->
    Errors.report_error ppf x;
    exit 2

let _ = Arg.parse arg_list arg_anon arg_usage

let () =
  if List.length !files_to_concat = 0 then begin
    prerr_endline "Usage: ./ocamlgcc [options] file1.ml [file2.ml...]";
    exit 2
  end

let main () =
  let files = List.rev !files_to_concat in
  let outputprefix = output_prefix (List.hd !files_to_concat) in
  let gcc = GccComp.build_gcc Format.err_formatter outputprefix files in
  dump_gcc outputprefix gcc

let () =
  try main ()
  with exn ->
    Errors.report_error Format.err_formatter exn;
    exit 1
