let array_of_map map =
  let h = list_length map in
  let w = match map with [] -> 0 | l :: _ -> list_length l in
  let amap = array_init h (fun _ -> array_create w Wall) in
  list_iteri
    (fun row y ->
       let arow = array_get amap y in
       list_iteri
         (fun cell x -> array_set arow x cell)
         row)
    map ;
  w, h, (fun x y -> array_get2 amap x y)

type state = direction * location list

let power_pill_value = 3
let pill_value = 1
let power_pill_bonus_value = 5
let pill_bonus_value = 5
let ghost_danger_value = 5
let ghost_danger_malus_value = 5
let group_value_limit_lo = 5
let group_value_limit_hi = 20

let ai_step world (ldir, lpath) =
  let w, h, cell = array_of_map world.map in
  let { x = sx ; y = sy } = world.lambda_man.location in
  let components =
    let marked, mark = marker w h  in
    let worklist = ref [ (sx, sy) ] in
    let components = ref [] in
    let rec complete elts neighs x y =
      match cell x y with
      | Wall -> ()
      | Pill | Power_pill as kind ->
          if not (marked x y) then begin
            mark x y ;
            elts := (x, y, kind) :: !elts ;
            complete elts neighs (x + 1) y ; complete elts neighs (x - 1) y ;
            complete elts neighs x (y + 1) ; complete elts neighs x (y - 1)
          end
        | _ -> neighs := (x, y) :: !neighs
    in
    let rec search x y =
      if not (marked x y) then
        match cell x y with
        | Wall -> ()
        | Pill | Power_pill -> collect x y
        | _ ->
            mark x y ;
            search (x + 1) y ; search (x - 1) y ;
            search x (y + 1) ; search x (y - 1)
    and collect x y =
      let elts = ref []
      and neighs = ref [] in
      complete elts neighs x y ;
      let rec value acc = function
        | [] -> acc
        | (_, _, Pill) :: tl -> value (acc + pill_value)  tl
        | (_, _, Power_pill) :: tl -> value (acc + power_pill_value)  tl
        | _ -> halt ()
      in
      let value =
        max group_value_limit_lo
          (min (value 0 !elts) group_value_limit_hi) in
      components := (value, !elts, !neighs) :: !components ;
      worklist := !neighs @ !worklist
    in
    while !worklist <> [] do
      let iteration = !worklist in
      worklist := [] ;
      list_iter (fun (x, y) -> search x y) iteration
    done ;
    !components
  in
  let apply_prio get set prio =
    let rec do_level = function
      | [] -> ()
      | (value, locs) :: tl ->
          match fill_level value locs with
          | [] -> do_level tl
          | neighs ->
              let prio = match tl with
                | [] -> [ (value - 1, neighs) ]
                | (value', neighs') as hd' :: tl' ->
                    if value - 1 = value' then
                      (value - 1, neighs @ neighs') :: tl'
                    else
                      (value - 1, neighs) :: tl
              in do_level prio
    and fill_level value locs =
      list_fold_left
        (fun r (x, y) ->
           if get x y <= value then begin
             set x y value ;
             let test x y l =
               if get x y <= value && cell x y <> Wall then (x, y) :: l else l in
             test x (y + 1) @@ test x (y - 1) @@
             test (x + 1) y @@ test (x - 1) y @@ r
           end else r)
        [] locs
    in do_level prio
  in
  let deli, set_deli =
    let compensation = w * h in
    let deli, set_deli = array w h 0 in  
    let prio = ref [] in
    let insert value contents =
      let rec rew = function
        | [] -> [ (value, contents) ]
        | (value', contents') as hd :: tl as l ->
            if value = value' then
              (value, contents @ contents') :: tl
            else if value < value' then
              hd :: rew tl
            else (* insert point *)
              (value, contents) :: l
      in
      prio := rew !prio
    in
    list_iter 
      (fun (value, elts, neighs) ->
         list_iter
           (fun (x, y, kind) ->
              match kind with
              | Power_pill | Pill -> set_deli x y (value + compensation)
              | _ -> halt ())
           elts ;
         insert (value - 1 + compensation) neighs)
      components ;
    apply_prio deli set_deli !prio ;
    list_iter 
      (fun (value, elts, neighs) ->
         list_iter
           (fun (x, y, kind) ->
              match kind with
              | Power_pill -> set_deli x y (deli x y + power_pill_bonus_value)
              | Pill -> set_deli x y (deli x y + pill_bonus_value)
              | _ -> halt ())
           elts)
      components ;
    (* debug_array deli w h ; *)
    deli, set_deli
  in
  let danger =
    let danger, set_danger = array w h 0 in
    let test x y l = if cell x y <> Wall then (x, y) :: l else l in
    let neighs =
      list_fold_left
        (fun tl { location = { x ; y }} ->
           set_danger x y ghost_danger_value ;
           test x (y + 1) @@ test x (y - 1) @@
           test (x + 1) y @@ test (x - 1) y @@ tl)
        [] world.ghosts
    in
    let prio = [ (ghost_danger_value - 1, neighs) ] in
    apply_prio danger
      (fun x y v -> set_danger x y v ; set_deli x y (deli x y - v))
      prio ;
    list_iter (* fix errors when ghosts are touching *)
      (fun { location = { x ; y }} ->
         set_danger x y (ghost_danger_value + ghost_danger_malus_value))
      world.ghosts ;
    (* debug_array danger w h ; *)
    danger
  in
  let eval_neighs x y eval =
    let test x y dir r =
      if cell x y <> Wall then list_insert (eval x y, dir) r else r in
    test x (y - 1) Up @@ test x (y + 1) Down @@
    test (x - 1) y Left @@ test (x + 1) y Right @@ []
  in
  let dir =
    match
      if danger sx sy > ghost_danger_value - 4 then
        eval_neighs sx sy (fun x y -> danger x y)
      else
        eval_neighs sx sy (fun x y -> -(deli x y))
    with
    | [] -> Up
    | [_, dir] -> dir
    | (v1, dir1) :: (v2, dir2) :: _ ->
        if v1 = v2 && dir1 = opp ldir then dir2 else dir1
  in
  (dir, world.lambda_man.location :: lpath), dir

let ai_init world =
  Down, [ world.lambda_man.location ]
