
(** Helpers *)

let rec retrieve_list retrieve_elem value =
  if atom value then []
  else retrieve_elem (car value) :: retrieve_list retrieve_elem (cdr value)

let retrieve_array retrieve_elem value =
  if atom value then [||]
  else
  let rec retrieve value offset =
    if atom (cdr value) then
      array_create (offset+1) (retrieve_elem (car value))
    else
      let arr = retrieve (cdr value) (succ offset) in
      arr.(offset) <- retrieve_elem (car value);
      arr in
  retrieve value 0

(** Map *)

type cell =
  | Wall
  | Empty
  | Pill
  | Power_pill
  | Fruit
  | Lambda_man
  | Ghost
let retrieve_cell (v : value) = (magic v : cell)

type row = cell list
let retrieve_row value = retrieve_array retrieve_cell value

type map = row list
let retrieve_map value = retrieve_array retrieve_row value

type location = { x : int; y : int }
let retrieve_location value =
  { x = magic (car value) ; y = magic (cdr value) }

type direction = Up | Right | Down | Left
let retrieve_direction value = (magic value : direction)

type lambda_man = {
  vitality : int; (* ticks before end of power pill activity *)
  location : location;
  direction : direction;
  lives : int;
  score : int;
}
let retrieve_lambda_man value =
  let vitality : int = magic (car value) in
  let location = retrieve_location (cadr value) in
  let direction = retrieve_direction (caddr value) in
  let lives : int = magic (cadddr value) in
  let score : int = magic (cddddr value) in
  { vitality ; location ; direction ; lives ; score }

type ghost_vitality =
  | Standard
  | Fright_mode
  | Invisible
let retrieve_ghost_vitality value = (magic value : ghost_vitality)

type ghost =
  { vitality : ghost_vitality;
    location : location;
    direction : direction }
let retrieve_ghost value =
  let vitality = retrieve_ghost_vitality (car value) in
  let location = retrieve_location (cadr value) in
  let direction = retrieve_direction (cddr value) in
  { vitality ; location ; direction }
type fruit = int option
let retrieve_fruit value =
  let v = (magic value : int) in
  if v = 0 then None else Some v

type world =
  { map : cell array array;
    lambda_man : lambda_man;
    ghosts : ghost list ;
    fruit : int option }
let retrieve_world value : world =
  let map = retrieve_map (car value) in
  let lambda_man = retrieve_lambda_man (cadr value) in
  let ghosts = retrieve_list retrieve_ghost (caddr value) in
  let fruit = retrieve_fruit (cdddr value) in
  { map ; lambda_man ; ghosts ; fruit }

let opp = function
  | Down -> Up
  | Left -> Right
  | Up -> Down
  | Right -> Left
